package Fedora::Rebuild;
use strict;
use warnings;
use version 0.77; our $VERSION = version->declare("v0.9.1");

use Moose;
use Moose::Util::TypeConstraints;
use File::Path;
use Carp;
use IO::Handle;
use Fedora::Rebuild::Types qw( Mode );
use Fedora::Rebuild::Execute;
use Fedora::Rebuild::Mock;
use Fedora::Rebuild::Package;
use Fedora::Rebuild::Repository;
use Fedora::Rebuild::Scheduler;
use Fedora::Rebuild::Set::Package;
use Fedora::Rebuild::Solver;
use namespace::clean;

has 'all' => (is => 'ro', isa => 'Str', required => 1);
has 'done' => (is => 'ro', isa => 'Str', required => 1);
has 'failed' => (is => 'ro', isa => 'Str', required => 1);
has 'workdir' => (is => 'ro', isa => 'Str', required => 1);
# Directory where repository with built packages live. This is needed for
# `mock' build mode.
has 'repodir' => (is => 'ro', isa => 'Str', required => 1);
# Reference to array of repository URLs. Usually points to Fedora mirror or
# Koji repository (baseurl in YUM terminology). This is needed for `mock'
# build mode.
has 'repourls' => (is => 'rw', isa => 'ArrayRef[Str]', required => 1);
# RPM architecture name, e.g. x86_64.
# TODO: Get current architecture by rpmGetArchInfo() from librpm
has 'architecture' => (is => 'ro', isa => 'Str', required => 1);
# Clone git repositories anonymously. This will not work with `koji' mode.
# Default is non-anonymous clone which requires preconfigured SSH public key
# authentication.
has 'anonymous' => (is => 'ro', isa => 'Bool', required => 0, default => 0);
# Git branch name
# "f14", "f15" etc. Use "rawhide" for latest one.
has 'dist' => (is => 'ro', isa => 'Str', required => 1);
# Fedpkg dist name identifier. This is required only if git branch name is not
# an offical one (e.g. a private branch).
# "f14", "f15" etc. Use "f18" for the latest one.
has 'pkgdist' => (is => 'ro', isa => 'Maybe[Str]', required => 0,
    default => undef);
# Build target name
# "dist-f14", "dist-f15" etc. Use "dist-rawhide" for latest one.
has 'target' => (is => 'ro', isa => 'Str', required => 1);
has 'message' => (is => 'ro', isa => 'Str', required => 1);
# Reference to function with three arguments (BuildRequire name, relation
# flag, version) returning false the BuildRequire should be considered, true
# otherwise. If attribute is undef or missing, no filtering will be performed
# (i.e. the same effect as sub {1}).
has 'buildrequiresfilter' => (is => 'ro', isa => 'CodeRef', required => 0);
# Pass "local" committing and building locally only, pass "koji" for pushing
# commits and building in Koji. Default is "koji" to build in Koji.
has 'mode' => (is => 'ro', isa => Mode, required => 0, default => 'koji');
# Run rebuild in given number of threads. Default is 1.
has 'threads' => (is => 'ro', isa => subtype('Int' => where {$_>0} =>
        message {"Attribute threads must be positive integer (was $_)"}),
    required => 0);
# Load binary provides of already done packages or buildrequires in given
# number of threads. This is to lower local I/O load. Default is 1. 
has 'loadthreads' => (is => 'ro', isa => subtype('Int' => where {$_>0} =>
        message {"Attribute loadthreads must be positive integer (was $_)"}),
    required => 0);
# Select rebuildable packages in given number of threads.
# Dependency solver is CPU intentsive work.
# Default is 1. 
has 'selectthreads' => (is => 'ro', isa => subtype('Int' => where {$_>0} =>
        message {"Attribute selectthreads must be positive integer (was $_)"}),
    required => 0);
# Maximal count of immediately sequential failures to accept and continue
# rebuild process. If the limit exceeds, rebuild process will terminate. This
# is good to catch pathologic cases when something is obviously wrong and
# should be fixed before rebuilding (e.g. Koji is down, or you have not SSH
# key loaded into SSH agent). Use non-positive number to disable this check.
# Default value is 0 (i.e. not to check).
has 'failurelimit' => (is => 'ro', isa => 'Int', required => 0, default => 0);
# Build packages in dependency order. Default is true.
# If set to false, the packages will be built irrespective to dependecies
# (build- and run-time).
has 'ordered' => (is => 'ro', isa => 'Bool', required => 0, default => 1);

has 'remaining_packages' => (is => 'ro', isa => 'Fedora::Rebuild::Set::Package',
    lazy_build => 1, init_arg => undef);
has 'done_packages' => (is => 'ro', isa => 'Fedora::Rebuild::Set::Package',
    lazy_build => 1, init_arg => undef);
has 'failed_packages' => (is => 'ro', isa => 'Fedora::Rebuild::Set::Package',
    lazy_build => 1, init_arg => undef);
has 'subsequent_failures' => (is => 'rw', isa => 'Int', default => 0,
    init_arg => undef);
has 'last_failed' => (is => 'rw', isa => 'Bool', default => 0,
    init_arg => undef);
has 'repository' => (is => 'ro', isa => 'Fedora::Rebuild::Repository',
    lazy_build => 1, init_arg => undef);


# Creates set of packages already rebuilt.
sub _build_done_packages {
    my $self = shift;
    my $packages = Fedora::Rebuild::Set::Package->new();
    my $file;

    if (not -e $self->done) {
        print "No packages have been rebuilt yet.\n";
        return $packages;
    }

    open($file, '<', $self->done) or
        croak "Could not open `" . $self->done .
            "' for loading list of already rebuilt packages: $!";
    print "Loading list of already rebuilt package names...\n";
    while (local $_ = <$file>) {
        chomp;
        if (m/^\s*$/) { next; }

        my $package = Fedora::Rebuild::Package->new(name => $_,
            workdir => $self->workdir, dist => $self->dist,
            pkgdist => $self->pkgdist, target => $self->target,
            message => $self->message);
        $packages->insert($package);
        $self->repository->insert($package);
    }
    if ($!) {
        croak "Could not read list of rebuilt package names from file `" .
        $self->done . "': $!";
    }
    close $file;
    $self->repository->update;
    print "Number of done packages: " . $packages->size() . "\n";
   
    return $packages;
}


sub mark_done {
    my ($self, $package) = @_;
    my $file = IO::Handle->new();
    
    open($file, '>>', $self->done) or
        croak "Could not open `" . $self->done . "' for appending: $!";
    printf $file $package->name . "\n" or
        croak "Could not write data into `" . $self->done . "': $!";
    $file->flush && $file->sync && close($file) or
        croak "Could not flush and close `" . $self->done . "': $!";

    $self->done_packages->insert($package);
    $self->last_failed(0);
    $self->subsequent_failures(0);
}

# Creates set of packages not yet rebuilt not already failed.
sub _build_remaining_packages {
    my $self = shift;
    my $packages = Fedora::Rebuild::Set::Package->new();
    my $file;

    open($file, '<', $self->all) or
        croak "Could not open " . $self->all .
            " for loading list of package names to rebuild: $!";
    print "Loading list of all package names to rebuild...\n";

    while (local $_ = <$file>) {
        chomp;
        if (m/^\s*$/) { next; }
        if ($packages->contains($_)) { next; }
        if (! $self->done_packages->contains($_) &&
                ! $self->failed_packages->contains($_)) {
            $packages->insert(Fedora::Rebuild::Package->new(name => $_,
                workdir => $self->workdir, dist => $self->dist,
                pkgdist => $self->pkgdist, target => $self->target,
                message => $self->message));
        }
    }
    if ($!) {
        croak "Could not read list of package names to rebuild from file `" .
            $self->all . "' :$!";
    }
    close $file;

    print "Number of all packages: " . 
        ($packages->size + $self->done_packages->size
            + $self->failed_packages->size) . "\n";
    print "Number of not yet rebuilt packages: " . $packages->size() . "\n";
    return $packages;
}


# Set of packages whose rebuild failed.
sub _build_failed_packages {
    my $self = shift;
    my $packages = Fedora::Rebuild::Set::Package->new();
    my $file;

    if (not -e $self->failed) {
        print "No packages have failed yet.\n";
        return $packages;
    }

    open($file, '<', $self->failed) or
        croak "Could not open `" . $self->failed .
            "' for loading list of already failed packages: $!";
    print "Loading list of already failed package names...\n";
    while (local $_ = <$file>) {
        chomp;
        if (m/^\s*$/) { next; }

        my $package = Fedora::Rebuild::Package->new(name => $_,
            workdir => $self->workdir, dist => $self->dist,
            pkgdist => $self->pkgdist, target => $self->target,
            message => $self->message);
        $packages->insert($package);
    }
    if ($!) {
        croak "Could not read list of failed package names from file `" .
        $self->failed . "': $!";
    }
    close $file;
    print "Number of failed packages: " . $packages->size() . "\n";
    
    return $packages;
}

sub _build_repository {
    my $self = shift;
    return Fedora::Rebuild::Repository->new(path => $self->repodir);
}

# Record package names into log of failed packages
sub mark_failed {
    my ($self, $package) = @_;
    my $file = IO::Handle->new();
    
    # Log failure
    open($file, '>>', $self->failed) or
        croak "Could not open `" . $self->failed . "' for appending: $!";
    printf $file $package->name . "\n" or
        croak "Could not write data into `" . $self->failed . "': $!";
    $file->flush && $file->sync && close($file) or
        croak "Could not flush and close `" . $self->failed . "': $!";

    # Move to list of failed
    $self->failed_packages->insert($package);

    # Check for very failures
    if ($self->last_failed) {
        $self->subsequent_failures($self->subsequent_failures + 1);
    } else {
        $self->last_failed(1);
        $self->subsequent_failures(1);
    }
    if ($self->failurelimit > 0 &&
            $self->subsequent_failures > $self->failurelimit) {
        croak "More then " . $self->failurelimit .
            " package(s) failed subsequently which is more then set " .
            "threshold. Aborting now.\n";
    }
}

# Load build-requires for each not-yet done package.
# Return true in case of success or croaks in case of failure.
sub load_sourcedependencies {
    my $self = shift;
    my @packages = $self->remaining_packages->packages;
    my $scheduler = Fedora::Rebuild::Scheduler->new(
        limit => $self->loadthreads,
        name => 'Loading build-requires',
        total => $#packages
    );
    my %jobs = ();
    my $i = 0;
    
    my $mock;
    if ($self->loadthreads == 1) {
        print "Initializing shared mock environment...\n";
        eval {
            $mock = Fedora::Rebuild::Mock->new(
                    architecture => $self->architecture,
                    repositories => $self->repourls,
                    shared => 1);
            $mock->make_configuration;
        };
        if ($@) {
            File::Path::remove_tree($mock->config_dir);
            croak("Could not configure mock: $@\n");
        }
        if (!Fedora::Rebuild::Execute->new->do(undef, 'mock',
                '--configdir', $mock->config_dir, '--root', $mock->config_root,
                '--no-cleanup-after', '--init')) {
            File::Path::remove_tree($mock->config_dir);
            croak("Could not initilize shared mock environment " .
                "for building source packages.\n");
        }
        print "Shared mock environment initialized successfully.\n";
    }

    print "Loading build-time dependenices of not yet rebuilt packages...\n";

    foreach my $package (@packages) {
        my $job = $scheduler->schedule($package->can('get_buildrequires'),
            $package, $self->mode, $self->repourls, $self->architecture,
            $self->anonymous, $mock);
        if (! defined $job) { next; }
        $jobs{$job} = $package;
        my %finished = $scheduler->finish(++$i < @packages);

        while (my ($job, $status) = each %finished) {
            my $package = $jobs{$job};
            if (!$$status[0]) {
                print "Could not load build-time dependencies for not yet " .
                    "built package `" . $package->name . "'.\n";
                print "Waiting for finishing scheduled jobs...\n";
                $scheduler->finish(1);
                print "All jobs loading build-time dependcies have finished.\n";
                if (defined $mock) {
                    File::Path::remove_tree($mock->config_dir);
                }
                croak "Could not load build-time dependencies.\n";
            }
        }
    }

    if (defined $mock) {
        File::Path::remove_tree($mock->config_dir);
    }
    print "Build-time dependencies of not-yet rebuilt packages loaded " .
        "successfully.\n";
    return 1;
}


# Load binary requires and provides of each done package.
# Return true in case of success or croaks in case of failure.
sub load_binarydependencies {
    my $self = shift;
    my @packages = $self->done_packages->packages;
    my $scheduler = Fedora::Rebuild::Scheduler->new(
        limit => $self->loadthreads,
        name => 'Loading binary dependendies',
        total => $#packages
    );
    my %jobs = ();
    my $i = 0;
    
    print "Loading binary dependenices of already rebuilt packages...\n";

    foreach my $package (@packages) {
        my $job = $scheduler->schedule($package->can('get_binarydependencies'),
            $package, $self->mode);
        if (! defined $job) { next; }
        $jobs{$job} = $package;
        my %finished = $scheduler->finish(++$i < @packages);

        while (my ($job, $status) = each %finished) {
            my $package = $jobs{$job};
            if (!$$status[0]) {
                print "Could not load binary dependencies for already built " .
                    "package `" . $package->name . "'.\n";
                print "Waiting for finishing scheduled jobs...\n";
                $scheduler->finish(1);
                print "All jobs loading binary dependcies have finished.\n";
                croak "Could not load binary dependencies\n";
            }
        }
    }

    print "Binary dependencies of already rebuilt packages loaded " .
        "successfully.\n";
    return 1;
}


# Decides a package is rebuildable now.
# Return 0 for false, 1 or true, undef for error while deciding.
sub is_rebuildable {
    my ($self, $package) = @_;
    my $is_rebuildable;
    my $message = '';

    if ($self->ordered) { 
        my $solver = Fedora::Rebuild::Solver->new(
            'packages' => $self->done_packages,
            'dependencyfilter' => $self->buildrequiresfilter);

        $is_rebuildable = $solver->is_buildable($package, \$message);
    } else {
        $is_rebuildable = 1;
        $message = "Package `" . $package->name .
            "' is buildable because unordered build mode is selected.";
    }

    $package->log_is_rebuildable($is_rebuildable, $message);
    print "$message\n";
    return $is_rebuildable;
}


# Return F:R:Set:Packages than can be rebuilt now
sub select_rebuildable {
    my $self = shift;
    my @packages = $self->remaining_packages->packages;
    my $scheduler = Fedora::Rebuild::Scheduler->new(
        limit => $self->selectthreads,
        name => 'Selecting buildable packages',
        total => $#packages
    );
    my %jobs = ();
    my %finished = ();
    print "Selecting rebuildable packages...\n";

    my $rebuildables = Fedora::Rebuild::Set::Package->new;
    foreach my $package (@packages) {
        my $job = $scheduler->schedule($self->can('is_rebuildable'), $self,
            $package);
        if (! defined $job) { next; }
        $jobs{$job} = $package;
        %finished = (%finished, ($scheduler->finished()));
    }
    %finished = (%finished, ($scheduler->finish()));

    for my $job (keys %finished) {
        my $job_status = $finished{$job};
        my $package = $jobs{$job};

        if ($$job_status[0]) {
            $rebuildables->insert($package);
        } elsif (!defined $job_status) {
            # Could not decide whether the $package is rebuildable. This is
            # fatal for the package. Move it to failed packages.
            $self->remaining_packages->delete($package);
            $self->mark_failed($package);
        }
    }

    print "Packages selected to rebuild (" . $rebuildables->size .
        "): " . $rebuildables->string . "\n";
    return $rebuildables;
}


# Rebuild all remaining packages
sub run {
    my $self = shift;
    if (defined $self->repository) {
        if (! defined $self->repository->url) {
            print "Starting repository HTTP server...\n";
            $self->repository->start;
            push @{$self->repourls}, $self->repository->url;
        }
        print "Repository URL is: <" . $self->repository->url . ">\n";
    }

    print "remaining_packages: " . $self->remaining_packages->string . "\n";
    print "done_packages: " . $self->done_packages->string . "\n";
    print "Rebuild mode: " . $self->mode .
        " " . (($self->ordered)?"ordered":"unordered") . "\n";

    if ($self->ordered) {
        $self->load_sourcedependencies;
        $self->load_binarydependencies;
    }

    while ($self->remaining_packages->size > 0) {
        my $rebuildable_packages = $self->select_rebuildable;
        if ($rebuildable_packages->size == 0) {
            printf "No more packages can be rebuilt!\n";
            last;
        }

        my $scheduler = Fedora::Rebuild::Scheduler->new(
            limit => $self->threads,
            name => 'Building',
            total => $rebuildable_packages->size
        );

        # Rebuild packages
        my %jobs = ();
        my @packages = $rebuildable_packages->packages;
        my $i = 0;
        
        foreach my $package (@packages) {
            $self->remaining_packages->delete($package);

            my $job = $scheduler->schedule($package->can('rebuild'), $package,
                $self->mode, $self->repourls, $self->architecture,
                $self->anonymous);
            if (! defined $job) { next; }
            $jobs{$job} = $package;
            my %finished = $scheduler->finish(++$i < @packages);

            while (my ($job, $status) = each %finished) {
                my $package = $jobs{$job};
                if ($$status[0]) {
                    # TODO: Push Provides into global list of available
                    # provides.
                    # XXX: Nothing here, fall through to rotation waiting,
                    # rebuilt package list tracked in $rebuildable_packages.
                } else {
                    $rebuildable_packages->delete($package);
                    $self->mark_failed($package);
                }
            }
        }


        # Wait for Koji rotation
        # This is separeted from rebuilding process to solve trade-off between
        # high threads number (consumes lot of memory and CPU time when
        # starting) and the rebuild itself is much faster Koji rotation period
        # (lot of time lost to wait for rotation).
        %jobs = ();
        @packages = $rebuildable_packages->packages;
        $i = 0;
        
        $scheduler = Fedora::Rebuild::Scheduler->new(
            limit => $self->threads,
            name => 'Waiting for build root',
            total => $rebuildable_packages->size
        );

        foreach my $package (@packages) {
            my $job = $scheduler->schedule($package->can('waitforbuildroot'),
                $package, $self->mode);
            if (! defined $job) { next; }
            $jobs{$job} = $package;
            my %finished = $scheduler->finish(++$i < @packages);

            while (my ($job, $status) = each %finished) {
                my $package = $jobs{$job};
                if ($$status[0]) {
                    $self->repository->insert($package);
                    $self->mark_done($package);
                } else {
                    $self->mark_failed($package);
                }
            }
        }
        $self->repository->update;
    }

    if ($self->remaining_packages->size > 0) {
        print "Following packages have unsatisfied build-time dependencies (" .
        $self->remaining_packages->size . "): " .
        $self->remaining_packages->string . "\n";
    }
    if ($self->failed_packages->size > 0) {
        print "Following packages have failed while rebuilding (" .
            $self->failed_packages->size . "): " .
            $self->failed_packages->string . "\n";
        return 0;
    } else {
        print "All rebuildable packages have been rebuilt.\n";
        return 1;
    }
}

1;
__END__
=encoding utf8

=head1 NAME

Fedora::Rebuild - Rebuilds Fedora packages from scratch

=head1 DESCRIPTION

Main goal is to rebuild perl modules packages for Fedora. The rebuild is
driven from bottom to top, i.e. from perl interpreter to modules depending on
intermediate modules. This way, it's possible to upgrade perl interpreter to
incompatible version and to rebuild all modules against the new interpreter.

Procedure is following: We have a set of source package to rebuild. We
distill all build-time dependencies for each source package. We ignore
non-perl dependencies as we focus on Perl only.

We select all source packages that do not depend on anything and push them
into Koji to rebuild. Obviously, the only fulfilling package is Perl interpret
only.

Once the first subset of packages is rebuilt, we gather their binary packages
and distill their binary provides. These freshly available provides are put
into new set of available provides. At the same time we remove the rebuilt
subset from original set of all source packages.

Then we wait for build root rotation in Koji to get rebuilt binary packages
available in build root.

(Of course we could get provides of new binary packages from koji repository
after rotation and we can get provides through repoquery instead of
inspecting binary packages manually.)

If package rebuild fails, the package is removed from future consideration.

Then we return to start of this procedure to select other subset of source
packages whose build-time dependecies can be satisfied from set of binary
provides we have obtained till now.

This loop cycles until set of source packages is empty.

=head1 SYNCHRONIZATION

Because mass rebuild is long term issue and lot of intermediate failures can
emerge, one must be able to resume failed rebuild process.

Safe resume is assured by proper logging. Once a package has been rebuilt, its
name is logged into list of done packages. Restarting rebuild program with the
same arguments loads done packages, remove them from set of remaining
packages, and populates list of available provides. In additon, packages whose
rebuild failed are removed from set of remaining packages.

=head1 AUTHOR

Petr Písař <ppisar@redhat.com>

=head1 COPYING

Copyright (C) 2011, 2012  Petr Pisar <ppisar@redhat.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

